from django.utils import six
from django.db.backends.base.operations import BaseDatabaseOperations


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = __name__.rsplit('.', 1)[0] + '.compiler'

    def quote_name(self, name):
        return name

    def last_insert_id(self, cursor, table_name, pk_name):
        return '{last_insert_id}'

    def random_function_sql(self):
        return 'RAND()'

    def sql_flush(self, style, tables, sequences, allow_cascade=False):
        # TODO: overwrite insert table by blank table.
        sql = ['%s %s;' % (
            style.SQL_KEYWORD('DROP TABLE'),
            style.SQL_FIELD(self.quote_name(table))
        ) for table in tables]

        return sql

    def value_to_db_datetime(self, value):
        if self.connection.features.supports_timezones:
            return super(DatabaseOperations, self).value_to_db_datetime(value)
        else:
            return six.text_type(value.replace(tzinfo=None))

    def start_transaction_sql(self):
        return ''

    def end_transaction_sql(self, success=True):
        return ''



from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.models.fields.related import ManyToManyField


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):

    sql_delete_table = "DROP TABLE %(table)s"
    sql_create_column = "ALTER TABLE %(table)s ADD COLUMNS (%(column)s %(definition)s)"
    sql_delete_column = "ALTER TABLE %(table)s DROP COLUMN %(column)s"

    def quote_value(self, value):
        from impala import dbapi

        if value is None:
            return 'NULL'
        elif isinstance(value, basestring):
            return "'%s'" % dbapi._escape(value)
        else:
            return str(value)

    def column_sql(self, model, field, include_default=False):
        db_params = field.db_parameters(connection=self.connection)
        sql = db_params['type']
        if sql is None:
            return None, None
        return sql, []

    def create_model(self, model):
        params = []
        column_sqls = []

        for field in model._meta.local_fields:

            definition, extra_params = self.column_sql(model, field)
            if definition is None:
                continue
            col_type_suffix = field.db_type_suffix(connection=self.connection)
            if col_type_suffix:
                definition += " %s" % col_type_suffix
            params.extend(extra_params)

            column_sqls.append("%s %s" % (
                self.quote_name(field.column),
                definition,
            ))

            if field.get_internal_type() == "AutoField":
                autoinc_sql = self.connection.ops.autoinc_sql(model._meta.db_table, field.column)
                if autoinc_sql:
                    self.deferred_sql.extend(autoinc_sql)

        sql = self.sql_create_table % {
            "table": self.quote_name(model._meta.db_table),
            "definition": ", ".join(column_sqls)
        }

        self.execute(sql, params)

        for field in model._meta.local_many_to_many:
            if field.rel.through._meta.auto_created:
                self.create_model(field.rel.through)

    def alter_unique_together(self, *args, **kwargs):
        pass

    def alter_index_together(self, *args, **kwargs):
        pass

    def add_field(self, model, field):
        if isinstance(field, ManyToManyField) and field.rel.through._meta.auto_created:
            return self.create_model(field.rel.through)

        definition, params = self.column_sql(model, field, include_default=True)

        if definition is None:
            return

        sql = self.sql_create_column % {
            "table": self.quote_name(model._meta.db_table),
            "column": self.quote_name(field.column),
            "definition": definition,
        }
        self.execute(sql, params)

        if self.connection.features.connection_persists_old_columns:
            self.connection.close()



import sys
import time
from django.utils.six.moves import input
from django.db.backends.base.creation import BaseDatabaseCreation


class DatabaseCreation(BaseDatabaseCreation):

    data_types = {
        'AutoField': 'bigint',  # TODO: string(uuid4)
        'BinaryField': 'string',
        'BigIntegerField': 'bigint',
        'BooleanField': 'boolean',
        'CharField': 'string',
        'CommaSeparatedIntegerField': 'string',
        'DateField': 'timestamp',
        'DateTimeField': 'timestamp',
        'DecimalField': 'decimal(%(max_digits)s, %(decimal_places)s)',
        'EmailField': 'string',
        'FileField': 'string',
        'FilePathField': 'string',
        'FloatField': 'float',
        'GenericIPAddressField': 'string',
        'ImageField': 'string',
        'IntegerField': 'bigint',
        'IPAddressField': 'string',
        'NullBooleanField': 'boolean',
        'OneToOneField': 'bigint',
        'PositiveIntegerField': 'bigint',
        'PositiveSmallIntegerField': 'int',
        'SlugField': 'string',
        'SmallIntegerField': 'smallint',
        'TextField': 'string',
        'TimeField': 'timestamp',
        'URLField': 'string',
        'XMLField': 'string',
    }

    data_type_check_constraints = {
        'PositiveIntegerField': '"%(column)s" >= 0',
        'PositiveSmallIntegerField': '"%(column)s" >= 0',
    }

    def sql_create_model(self, model, style, known_models=set()):
        opts = model._meta
        if not opts.managed or opts.proxy or opts.swapped:
            return [], {}

        final_output = []
        table_output = []
        pending_references = {}
        qn = self.connection.ops.quote_name

        for f in opts.local_fields:
            col_type = f.db_type(connection=self.connection)
            if col_type is None:
                continue

            field_output = [
                style.SQL_FIELD(qn(f.column)),
                style.SQL_COLTYPE(col_type)
            ]

            table_output.append(' '.join(field_output))

        full_statement = [
            style.SQL_KEYWORD('CREATE TABLE') + ' ' +
            style.SQL_TABLE(qn(opts.db_table)) + ' ('
        ]

        for i, line in enumerate(table_output):
            full_statement.append(
                '    %s%s' % (line, ',' if i < len(table_output) - 1 else ''))

        full_statement.append(')')

        # TODO: hive table space options

        full_statement.append(';')
        final_output.append('\n'.join(full_statement))

        return final_output, pending_references

    def _create_test_db(self, verbosity, autoclobber):
        qn = self.connection.ops.quote_name
        test_database_name = self._get_test_db_name()

        # Create the test database and connect to it.
        with self._nodb_connection.cursor() as cursor:
            try:
                cursor.execute("CREATE DATABASE %s" % qn(test_database_name))
            except Exception as e:
                sys.stderr.write(
                    "Got an error creating the test database: %s\n" % e)

                if not autoclobber:
                    confirm = input(
                        "Type 'yes' if you would like to try deleting the test "
                        "database '%s', or 'no' to cancel: " % test_database_name)

                if autoclobber or confirm == 'yes':
                    try:
                        if verbosity >= 1:
                            print("Destroying old test database '%s'..." %
                                  self.connection.alias)

                        self._destroy_test_db(test_database_name, verbosity)
                        cursor.execute("CREATE DATABASE %s" % qn(test_database_name))
                    except Exception as e:
                        sys.stderr.write(
                            "Got an error recreating the test database: %s\n" % e)
                        sys.exit(2)
                else:
                    print("Tests cancelled.")
                    sys.exit(1)

        return test_database_name

    def _destroy_test_db(self, test_database_name, verbosity):
        qn = self.connection.ops.quote_name

        with self._nodb_connection.cursor() as cursor:
            time.sleep(1)

            cursor.execute("use %s" % qn(test_database_name))
            cursor.execute('show tables')

            for name in cursor.fetchall():
                cursor.execute('DROP TABLE %s' % name[0])

            cursor.execute("use default")
            cursor.execute(
                "DROP DATABASE %s" % qn(test_database_name))

    def sql_indexes_for_model(self, model, style):
        return []

    def set_autocommit(self):
        pass



import uuid
from django.db.models.sql import compiler
from django.db.models.fields import AutoField


class SQLCompiler(compiler.SQLCompiler):
    def get_default_columns(self, *args, **kwargs):
        results = super().get_default_columns(*args, **kwargs)
        autos = []
        for col in results:
            if isinstance(col.field, AutoField):
                autos.append(col)
        for auto in autos:
            results.remove(col)
        return results
                


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):

    _latest_param = None

    def as_sql(self):
        opts = self.query.get_meta()
        origfields = self.query.fields

        if opts.has_auto_field:
            self.query.fields = [opts.pk] + self.query.fields

        queries = []
        for sql, params in super(SQLInsertCompiler, self).as_sql():
            if opts.has_auto_field:
                params[0] = int(str(int(uuid.uuid4()))[:18])  # TODO: string(uuid4)

            self._latest_param = params
            queries.append((sql, params))

        self.query.fields = origfields
        return queries

    def execute_sql(self, return_id=False):
        result = (
            super(SQLInsertCompiler, self)
            .execute_sql(return_id=return_id))

        if not result:
            return

        return result.format(
            last_insert_id=self._latest_param[0])


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    pass


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    pass


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    pass


# class SQLDateCompiler(compiler.SQLDateCompiler, SQLCompiler):
#     pass


# class SQLDateTimeCompiler(compiler.SQLDateTimeCompiler, SQLCompiler):
#     pass



import os
import sys

from django.db.backends.base.client import BaseDatabaseClient


class DatabaseClient(BaseDatabaseClient):
    executable_name = 'impala-shell'

    def runshell(self):
        settings_dict = self.connection.settings_dict
        args = [self.executable_name]
        db = settings_dict['OPTIONS'].get('db', settings_dict['NAME'])
        user = settings_dict['OPTIONS'].get('user', settings_dict['USER'])
        passwd = settings_dict['OPTIONS'].get('passwd', settings_dict['PASSWORD'])
        host = settings_dict['OPTIONS'].get('host', settings_dict['HOST'])
        port = settings_dict['OPTIONS'].get('port', settings_dict['PORT'])

        if user:
            args += ["--user=%s" % user]
        if passwd:
            args += ["--password=%s" % passwd]
        if host:
            args += ["--host=%s" % host]
        if port:
            args += ["--port=%s" % port]
        if db:
            args += [db]

        if os.name == 'nt':
            sys.exit(os.system(" ".join(args)))
        else:
            os.execvp(self.executable_name, args)
